# @author Augustin Mortier
# @desc A-Profiles - AeronetsData class


class AeronetData:
    """
    Base class representing profiles data returned by (aprofiles.io.reader.ReadAeronet):."""

    def __init__(self, data):
        self.data = data
        raise NotImplementedError("AeronetData is not implemented yet")

def _main():
    pass

if __name__ == "__main__":
    _main()



# Some scripts to be run when aprofiles is loaded

from . import aeronet, mec, profiles, rayleigh, reader, size_distribution, utils
from .detection import clouds, foc, pbl
from .io import read_aeronet, read_eprofile, write_profiles
from .plot import image, profile
from .retrieval import extinction, mass_conc, ref_altitude
from .simulation import ext2back



# @author Augustin Mortier
# @desc A-Profiles - Code for creating map json file to be used in V-Profiles

import json
import warnings
from pathlib import Path

import numpy as np
import xarray as xr


def make_map(path, yyyy: str, mm: str, mapname) -> None:
    # one map, per day, which collects the maximum extinction with no low-level clouds (<6km) at each station
    with open(Path(path, yyyy, mm, mapname), 'w') as json_file:
        json.dump({}, json_file)

def add_to_map(fn, path, yyyy: str, mm: str, dd: str, mapname) -> None:
    # map collects the maximum extinction value with no low-level clouds (<6km) at each station at a hourly resolution
    # for each station, write an array with extinction values, and array with scenes for each hour of the day
    
    # read data
    vars_to_read = ['extinction', 'retrieval_scene', 'cloud_amount', 'lidar_ratio', 'aer_type', 'mec']
    
    # in some cases, the file might be corrupted. Just skip it then.
    try:
        ds = xr.open_dataset(fn, engine='netcdf4', chunks=-1)[vars_to_read].load()
    except OSError:
        print(f'File {fn} is corrupted. Skipping...')
        return

    # calculate the max extinction and determine the scene for each hour of the day
    # need to convert time (from int to datetime) in order to use the resample method from xarray
    ds["time"]= ds.time.data.astype("datetime64[ms]").astype("datetime64[ns]")

    # in order to prevent some monotony issue, sort by time
    ds = ds.sortby('time')

    layers = ['0-6', '0-2', '2-4', '4-6']
    max_ext = {}
    for layer in layers:
        zmin = float(layer.split('-')[0])*1000
        zmax = float(layer.split('-')[1])*1000
        zmin_asl = zmin + ds.attrs['station_altitude']
        zmax_asl = zmax + ds.attrs['station_altitude']
        
        # extinction for each hour
        max_ext_profiles = ds.where((ds.altitude>=zmin_asl) & (ds.altitude<zmax_asl), drop=True).extinction.resample(time="1h").max().data

        # take the maximum value in each profile
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            max_ext[layer] = np.nanmax(max_ext_profiles, axis=1)

    # scene for each hour
    max_retrieval_scene = ds.retrieval_scene.resample(time='1h').max().data
    max_cloud_amount = ds.cloud_amount.resample(time='1h').max().data
    mean_lidar_ratio = ds.lidar_ratio.resample(time='1h').mean().data
    
    # make mec dictionary
    mec = dict()
    for i, aer_type in enumerate(ds.aer_type.data):
        mec[aer_type] = round(float(ds.mec[i].data), 3)

    # open current map
    with open(Path(path, yyyy, mm, mapname), 'r') as json_file:
        data = json.load(json_file)
    json_file.close()        

    # add new data to calendar data
    station_id = f'{ds.attrs["wigos_station_id"]}-{ds.attrs["instrument_id"]}'
    if station_id not in data:
        data[station_id] = {}
    data[station_id][dd] = {
        'attrs': {
            'coords': {
                'latitude': ds.attrs['station_latitude'],
                'longitude': ds.attrs['station_longitude'],
                'altitude': ds.attrs['station_altitude'],
            },
            'l0_wavelength': ds.attrs['l0_wavelength'],
            'station_id': station_id,
            'station_name': ds.attrs['site_location'],
            'instrument_type': ds.attrs['instrument_type']
        },
        'max_ext:0-6km': [round(ext,4) if not np.isnan(ext) else None for ext in max_ext['0-6']],
        'max_ext:0-2km': [round(ext,4) if not np.isnan(ext) else None for ext in max_ext['0-2']],
        'max_ext:2-4km': [round(ext,4) if not np.isnan(ext) else None for ext in max_ext['2-4']],
        'max_ext:4-6km': [round(ext,4) if not np.isnan(ext) else None for ext in max_ext['4-6']],
        'retrieval_scene': [retrieval_scene if not np.isnan(retrieval_scene) else None for retrieval_scene in max_retrieval_scene.tolist()],
        'cloud_amount': [cloud_amount if not np.isnan(cloud_amount) else None for cloud_amount in max_cloud_amount.tolist()],
        'lidar_ratio': [lidar_ratio if not np.isnan(lidar_ratio) else None for lidar_ratio in mean_lidar_ratio.tolist()],
        'mec': mec
    }

    # write new map
    with open(Path(path, yyyy, mm, mapname), 'w') as json_file:
        json.dump(data, json_file)



# @author Augustin Mortier
# @desc A-Profiles - Size Distribution

import json

import matplotlib.pyplot as plt
import numpy as np

from pathlib import Path

from aprofiles import utils


class SizeDistributionData:
    """
    Class for computing *size distributions* for a given aerosol type.
    This class calls the `aprofiles.size_distribution.SizeDistributionData.get_sd`: method, which calculates VSD (Volume Size Distribution) and NSD (Number Size Distribution).

    Attributes:
        `aer_type` ({'dust', 'volcanic_ash', 'biomass_burning','urban'}): aerosol type.
        `aer_properties` (dict): dictionnary describing the optical and microphophysical properties of the prescribed aerosol (read from *aer_properties.json*)

    Example:
        ```python
        # some imports
        import aprofiles as apro
        sd = apro.size_distribution.SizeDistributionData('urban')
        # checkout the instance attributes
        apro.size_distribution.SizeDistributionData('dust').__dict__.keys()
        dict_keys(['aer_type', 'aer_properties', 'radius', 'vsd', 'nsd'])
        ```
    """

    def __init__(self, aer_type):
        self.aer_type = aer_type

        # read aer_properties.json files
        f = open(Path(Path(__file__).parent,'config','aer_properties.json'))
        aer_properties = json.load(f)
        f.close()
        # check if the aer_type exist in the json file
        if aer_type not in aer_properties.keys():
            raise ValueError(
                f"{aer_type} not found in aer_properties.json. `aer_type` must be one of the follwowing: {list(aer_properties.keys())}"
            )
        else:
            self.aer_properties = aer_properties[self.aer_type]
            self.get_sd()

    def _vsd_to_nsd(self):
        """
        Transforms Volume Size Distribution to Number Size Distribution
        """
        self.nsd = [
            self.vsd[i] * 3 / (4 * np.pi * self.radius[i] ** 4)
            for i in range(len(self.radius))
        ]
        return self

    def get_sd(self):
        """
        Returns the Volume and Number Size Distributions arrays from an instance of the (SizeDistributionData): class .

        Returns:
            (SizeDistribData): object with additional attributes:

                - `radius` (numpy.ndarray): radius (µm)
                - `vsd` (numpy.ndarray): Volume Size Distribution (µm2.µm-3)
                - `nsd` (numpy.ndarray): Number Size Distribution (µm-3.µm-1)
        """

        aer_properties = self.aer_properties
        radius = np.arange(1e-2, 2e1, 1e-3)  # radius in µm
        vsd = np.zeros(len(radius))

        # we loop though all the keys defining the different modes
        for mode in aer_properties["vsd"].keys():
            vsd += utils.gaussian(
                np.log(radius),
                np.log(aer_properties["vsd"][mode]["reff"]),
                aer_properties["vsd"][mode]["rstd"],
                aer_properties["vsd"][mode]["conc"],
            )

        self.radius = radius
        self.vsd = vsd
        self._vsd_to_nsd()

        return self

    def plot(self, show_fig=True):
        """
        Plot Size Distributions of an instance of the (SizeDistributionData): class.

        Example:
            ```python
            # import aprofiles
            import aprofiles as apro
            # get size distribution for urban particles
            sd = apro.size_distribution.SizeDistribData('urban');
            # plot profile
            sd.plot()
            ```

            ![Size distributions for urban particles](../../assets/images/urban_sd.png)
        """
        fig, ax = plt.subplots(1, 1, figsize=(6, 6))

        # plot Volume Size Distribution in 1st axis
        print(self.vsd)
        ax.plot(self.radius, self.vsd, label="VSD")
        ax.set_ylabel("V(r) (µm2.µm-3)")

        # plot Number Size Distribution in 2nd axis
        if "nsd" in self.__dict__:
            # add secondary yaxis
            ax2 = ax.twinx()
            ax2.plot(self.radius, self.nsd, "orange", label="NSD")
            ax2.set_ylabel("N(r) (µm-3.µm-1)")
            # ax2.set_ylim([0,10])

        ax.set_xlabel("Radius (µm)")
        ax.set_xscale("log")
        fig.legend(
            loc="upper right", bbox_to_anchor=(1, 1), bbox_transform=ax.transAxes
        )
        plt.title(
            f"Size Distribution for {self.aer_type.capitalize().replace('_', ' ')} Particles",
            weight="bold",
        )
        plt.tight_layout()
        if show_fig:
            plt.show()


def _main():
    import aprofiles as apro
    sd_data = SizeDistributionData("urban")
    sd_data.plot()


if __name__ == "__main__":
    _main()



    # @author Augustin Mortier
# @desc A-Profiles - Code for creating climatology json file to be used in V-Profiles

from datetime import datetime
from pathlib import Path
import os

import numpy as np
import orjson
import xarray as xr


def compute_climatology(path, station_id, season_variables, all_variables, aerosols_only):
    # get all files
    station_files = []
    for root, dirs, files in os.walk(path, followlinks=True):
        for file in files:
            if station_id in file and file.endswith(".nc"):
                station_files.append(os.path.join(root, file))
    
    try:
        # open dataset with xarray
        vars = season_variables + all_variables + ['retrieval_scene', 'cloud_amount', 'scene']
        ds = xr.open_mfdataset(station_files, parallel=False, decode_times=True, chunks=-1, combine='nested', compat='override')[vars].load()
        
        # store attributes which are destroyed by the resampling method
        attrs = ds.attrs
        # replace np.int by int
        for attr in attrs:
            if isinstance(attrs[attr], np.uint32):
                attrs[attr] = int(attrs[attr])
        
        # keep only clear scenes
        if aerosols_only:
            ds = ds.where((ds.retrieval_scene <= 1) & (ds.cloud_amount == 0))
            
        # seasonal resampling
        Qds = ds.resample(time="QE").mean().compute()

        # add some statistics
        attrs["ndays"] = {"ndays": len(station_files), "since": str(ds.time.data[0]).split("T")[0]}
        attrs["today"] = datetime.today().strftime("%Y-%m-%d")

        # add number of days per season as a new variable
        Qds["ndays"] = ds.scene.resample(time="D").count().resample(time="QE").count()

        # add Z to time
        ds.coords['time'] = (ds['time'].astype(int)  * 1e-6).astype(int)
        Qds.coords['time'] = (Qds['time'].astype(int)  * 1e-6).astype(int)

        # round altitude to 3 decimals
        ds.coords['altitude'] = ds['altitude'].round(3)
        Qds.coords['altitude'] = Qds['altitude'].round(3)

        # select variables
        multivars_dict = {}

        # add seasonal variables
        for var in season_variables + ["ndays"]:
            multivars_dict[var] = Qds[var].round(6).to_dict()
        
        # add all variables
        for var in all_variables:
            multivars_dict[var] = ds[var].round(6).to_dict()

        # add attributes as separate key
        multivars_dict["attrs"] = attrs

        # define path
        clim_path = Path(path, "climato")
        # create directory if does not exist
        clim_path.mkdir(parents=True, exist_ok=True)

        # write data to json file
        with open(Path(clim_path, f"AP_{station_id}_clim.json"), 'wb') as json_file:
            json_file.write(orjson.dumps(multivars_dict, option=orjson.OPT_SERIALIZE_NUMPY))

    except ValueError:
        print(f'ValueError encountered with {station_id}')



# @author Augustin Mortier
# @desc A-Profiles - E-PROFILE reading class

import numpy as np
import xarray as xr


class ReadEPROFILE:
    """
    E-PROFILE reading class.

    Data:
        [https://data.ceda.ac.uk/badc/eprofile/data/daily_files/]

    Attributes:
        path (str): The path of the file to be read (path or URL).
    """

    def __init__(self, path):
        self.path = path

    def _check_path(self):
        """
        Check the path

        Raises:
            OSError: [description]
            NotImplementedError: [description]
        """
        # check if file is NetCDF
        filename = self.path.split("/")[-1]
        if not filename.endswith(".nc"):
            raise OSError(f"NetCDF file required (.nc). File format not supported {filename}")
        # check if web address
        if self.path.startswith("https://"):
            raise NotImplementedError(
                "The reading of CEDA Archive is not implemented yet"
            )

    def _read_l2file(self):
        """
        Read single file using `xr.open_dataset()`.

        Returns:
            (xarray.Dataset):
        """
        ds = xr.open_dataset(self.path, decode_times=True)
        # remove time duplicate values if exists
        if len(ds.time.values) != len(np.unique(ds.time.values)):
            ds = ds.drop_duplicates(dim='time')
        # in CEDA archive, dimensions come as (altitude, time). Transpose all variables which have altitude as dimension.
        if ds.latitude.dims[0]=='altitude':
            ds = ds.transpose(..., 'altitude')
        # replace wavelength with actual value in attenuated backscatter longname
        ds.attenuated_backscatter_0.attrs["long_name"] = ds.attenuated_backscatter_0.long_name.replace(
            "at wavelength 0", f"@ {int(ds.l0_wavelength.data)} nm"
        )
        ds.attenuated_backscatter_0.attrs["units"] = ds.attenuated_backscatter_0.attrs["units"].replace(
            "1E-6*1/(m*sr)", ("Mm-1.sr-1")
        )
        return ds

    def read(self):
        """
        Method which reads E-PROFILE data.

        Returns:
            (xarray.Dataset):
        """
        
        self._check_path()
        
        # check if the filename starts with L2
        filename = self.path.split("/")[-1]
        if filename.startswith("L2"):
            return self._read_l2file()
        else:
            raise NotImplementedError("Only L2 files are supported at the moment")



# @author Augustin Mortier
# @desc A-Profiles - ProfilesData class

import copy
from pathlib import Path

import numpy as np
import xarray as xr
from rich.progress import track
from typing import Literal

import aprofiles as apro


class ProfilesData:
    """
    Base class representing profiles data returned by (aprofiles.reader.ReadProfiles):.
    """

    def __init__(self, data):
        self.data = data

    @property
    def data(self):
        """
        Data attribute (instance of (xarray.Dataset):)
        """
        return self._data

    @data.setter
    def data(self, data):
        if not isinstance(data, xr.Dataset):
            raise ValueError("Wrong data type: an xarray Dataset is expected.")
        self._data = data

    def _get_index_from_altitude_AGL(self, altitude):
        """
        Returns the closest index of the ProfilesData vertical dimension to a given AGL altitude

        Args:
            altitude (float): in m, altitude AGL to look for

        Returns:
            (int): Closest index of the vertical dimension to the given altitude AGL
        """
        altitude_asl = altitude + self.data.station_altitude.data
        return int(np.argmin(abs(self.data.altitude.data - altitude_asl)))

    def _get_resolution(self, which):
        """
        Returns the resolution of a given dimension. Supports 'altitude' and 'time'.
        The altitude resolution is given in meters, while the time resolution is given in seconds.

        Args:
            which (str): Dimension: must be ['altitude', 'time'].

        Returns:
            (float): resolution, in m (if which=='altitude') or in s (if which=='time')
        """
        if which == "altitude":
            return min(np.diff(self.data.altitude.data))
        elif which == "time":
            return (
                min(np.diff(self.data.time.data)).astype("timedelta64[s]").astype(int)
            )
    
    def _get_lowest_clouds(self):
        # returns an array of the altitude (in m, ASL) of the lowest cloud for each timestamp
        lowest_clouds = np.full(np.shape(self.data.time.data), np.nan)
        for i, _ in enumerate(self.data.time.data):
            clouds_profile = self.data.clouds[i,:]
            i_clouds = np.squeeze(np.where(clouds_profile))
            if len(i_clouds) > 0:
                lowest_clouds[i] = self.data.altitude.data[i_clouds[0]]
        return lowest_clouds

    def _get_itime(self, datetime):
        """
        Returns the index of closest datetime available of the ProfilesData object.
        """
        time = self.data.time.data
        i_time = np.argmin(abs(time - datetime))
        return i_time

    def snr(self, var="attenuated_backscatter_0", step=4, verbose=False):
        """
        Method that calculates the Signal to Noise Ratio.

        Args:
            var (str, optional): Variable of the DataArray to calculate the SNR from.
            step (int, optional): Number of steps around we calculate the SNR for a given altitude.
            verbose (bool, optional): Verbose mode.

        Returns:
            (ProfilesData): object with additional (xarray.DataArray): `snr`.

        Example:
            ```python
            import aprofiles as apro
            # read example file
            path = "examples/data/E-PROFILE/L2_0-20000-001492_A20210909.nc"
            reader = apro.reader.ReadProfiles(path)
            profiles = reader.read()
            # snr calculation
            profiles.snr()
            # snr image
            profiles.plot(var='snr',vmin=0, vmax=3, cmap='Greys_r')
            ```
        """
        
        # Get the dimensions of the data array
        time_len, altitude_len = self.data[var].shape
        
        # Preallocate the SNR array with zeros
        snr_array = np.zeros((time_len, altitude_len))
        
        # Loop over each time step
        for t in track(range(time_len), description="snr   ", disable=not verbose):
            # Extract 1D slice for current time step
            array = self.data[var].data[t, :]
            
            # Create a sliding window view for the rolling calculation
            sliding_windows = np.lib.stride_tricks.sliding_window_view(array, window_shape=2*step+1, axis=0)
            
            # Calculate mean and std across the window axis
            means = np.nanmean(sliding_windows, axis=1)
            stds = np.nanstd(sliding_windows, axis=1)
            
            # Handle the edges (where sliding window can't be applied due to boundary)
            means = np.pad(means, pad_width=(step,), mode='constant', constant_values=np.nan)
            stds = np.pad(stds, pad_width=(step,), mode='constant', constant_values=np.nan)
            
            # Avoid division by zero
            stds = np.where(stds == 0, np.nan, stds)
            
            # Compute SNR
            snr_array[t, :] = np.divide(means, stds, where=(stds != 0))
        
        # Add the SNR DataArray to the object's data attribute
        self.data["snr"] = (('time', 'altitude'), snr_array)
        self.data["snr"] = self.data.snr.assign_attrs({
            'long_name': 'Signal to Noise Ratio',
            'units': '',
            'step': step
        })

        return self

    def gaussian_filter(
        self, sigma=0.25, var="attenuated_backscatter_0", inplace=False
    ):
        """
        Applies a 2D gaussian filter in order to reduce high frequency noise.

        Args:
            sigma (scalar or sequence of scalars, optional): Standard deviation for Gaussian kernel. The standard deviations of the Gaussian filter are given for each axis as a sequence, or as a single number, in which case it is equal for all axes.
            var (str, optional): variable name of the Dataset to be processed.
            inplace (bool, optional): if True, replace the instance of the (ProfilesData): class.

        Returns:
            (ProfilesData): object with additional attributes `gaussian_filter` for the processed (xarray.DataArray):.

        Example:
            ```python
            import aprofiles as apro
            # read example file
            path = "examples/data/E-PROFILE/L2_0-20000-001492_A20210909.nc"
            reader = apro.reader.ReadProfiles(path)
            profiles = reader.read()
            # apply gaussian filtering
            profiles.gaussian_filter(sigma=0.5, inplace=True)
            profiles.data.attenuated_backscatter_0.attrs.gaussian_filter
            0.50
            ```

            ![Before gaussian filtering](../../assets/images/attenuated_backscatter.png)
            ![After gaussian filtering (sigma=0.5)](../../assets/images/gaussian_filter.png)
        """
        import copy

        from scipy.ndimage import gaussian_filter

        # apply gaussian filter
        filtered_data = gaussian_filter(self.data[var].data, sigma=sigma)

        if inplace:
            self.data[var].data = filtered_data
            new_dataset = self
        else:
            copied_dataset = copy.deepcopy(self)
            copied_dataset.data[var].data = filtered_data
            new_dataset = copied_dataset
        # add attribute
        new_dataset.data[var].attrs["gaussian_filter"] = sigma
        return new_dataset

    def time_avg(self, minutes, var="attenuated_backscatter_0", inplace=False):
        """
        Rolling median in the time dimension.

        Args:
            minutes (float): Number of minutes to average over.
            var (str, optional): variable of the Dataset to be processed.
            inplace (bool, optional): if True, replace the instance of the (ProfilesData): class.

        Returns:
            (ProfilesData):
        """
        rcs = self.data[var].copy()
        # time conversion from minutes to seconds
        t_avg = minutes * 60
        # time resolution in profiles data in seconds
        dt_s = self._get_resolution("time")
        # number of timestamps to be to averaged
        nt_avg = max([1, round(t_avg / dt_s)])
        # rolling median
        filtered_data = (
            rcs.rolling(time=nt_avg, min_periods=1, center=True).median().data
        )

        if inplace:
            self.data[var].data = filtered_data
            new_dataset = self
        else:
            copied_dataset = copy.deepcopy(self)
            copied_dataset.data[var].data = filtered_data
            new_dataset = copied_dataset
        # add attribute
        new_dataset.data[var].attrs["time averaged (minutes)"] = minutes
        return new_dataset

    def extrapolate_below(
        self, var="attenuated_backscatter_0", z=150, method="cst", inplace=False
    ):
        """
        Method for extrapolating lowest layers below a certain altitude. This is of particular intrest for instruments subject to After Pulse effect, with saturated signal in the lowest layers.
        We recommend to use a value of zmin=150m due to random values often found below that altitude which perturbs the clouds detection.

        Args:
            var (str, optional): variable of the :class:`xarray.Dataset` to be processed.
            z (float, optional): Altitude (in m, AGL) below which the signal is extrapolated.
            method ({'cst', 'lin'}, optional): Method to be used for extrapolation of lowest layers.
            inplace (bool, optional): if True, replace the instance of the (ProfilesData): class.

        Returns:
            (ProfilesData): object with additional attributes:
            
                - `extrapolation_low_layers_altitude_agl`
                - `extrapolation_low_layers_method` for the processed (xarray.DataArray):.

        Example:
            ```python
            import aprofiles as apro
            # read example file
            path = "examples/data/E-PROFILE/L2_0-20000-001492_A20210909.nc"
            reader = apro.reader.ReadProfiles(path)
            profiles = reader.read()
            # desaturation below 4000m
            profiles.extrapolate_below(z=150., inplace=True)
            profiles.data.attenuated_backscatter_0.extrapolation_low_layers_altitude_agl
            150
            ```

            ![Before extrapolation](../../assets/images/lowest.png)
            ![After extrapolation](../../assets/images/lowest_extrap.png)
        """

        # get index of z
        imax = self._get_index_from_altitude_AGL(z)

        nt = np.shape(self.data[var].data)[0]

        if method == "cst":
            # get values at imin
            data_zmax = self.data[var].data[:, imax]
            # generates ones matrice with time/altitude dimension to fill up bottom
            ones = np.ones((nt, imax))
            # replace values
            filling_matrice = np.transpose(np.multiply(np.transpose(ones), data_zmax))
        elif method == "lin":
            raise NotImplementedError("Linear extrapolation is not implemented yet")
        else:
            raise ValueError("Expected string: lin or cst")

        if inplace:
            self.data[var].data[:, 0:imax] = filling_matrice
            new_profiles_data = self
        else:
            copied_dataset = copy.deepcopy(self)
            copied_dataset.data[var].data[:, 0:imax] = filling_matrice
            new_profiles_data = copied_dataset

        # add attributes
        new_profiles_data.data[var].attrs["extrapolation_low_layers_altitude_agl"] = z
        new_profiles_data.data[var].attrs["extrapolation_low_layers_method"] = method
        return new_profiles_data

    def range_correction(self, var="attenuated_backscatter_0", inplace=False):
        """
        Method that corrects the solid angle effect (1/z2) which makes that the backscatter beam is more unlikely to be detected with the square of the altitude.

        Args:
            var (str, optional): variable of the Dataset to be processed.
            inplace (bool, optional): if True, replace the instance of the (ProfilesData): class.

        Returns:
            (ProfilesData):

        .. warning::
            Make sure that the range correction is not already applied to the selected variable.
        """

        # for the altitude correction, must one use the altitude above the ground level
        z_agl = self.data.altitude.data - self.data.station_altitude.data

        data = self.data[var].data
        range_corrected_data = np.multiply(data, z_agl)

        if inplace:
            self.data[var].data = range_corrected_data
            new_profiles_data = self
        else:
            copied_dataset = copy.deepcopy(self)
            copied_dataset.data[var].data = range_corrected_data
            new_profiles_data = copied_dataset

        # add attribute
        new_profiles_data.data[var].attrs["range correction"] = True
        # remove units
        new_profiles_data.data[var].attrs["units"] = None
        return new_profiles_data

    def desaturate_below(self, var="attenuated_backscatter_0", z=4000.0, inplace=False):
        """
        Remove saturation caused by clouds at low altitude which results in negative values above the maximum.
        The absolute value of the signal is returned below the prescribed altitude.

        Args:
            var (str, optional): variable of the :class:`xarray.Dataset` to be processed.
            z (float, optional): Altitude (in m, AGL) below which the signal is unsaturated.
            inplace (bool, optional): if True, replace the instance of the (ProfilesData):.

        Todo:
            Refine method to desaturate only saturated areas.

        Returns:
            (ProfilesData): object with additional attribute `desaturate` for the processed (xarray.DataArray):.

        .. warning::
            For now, absolute values are returned everywhere below the prescribed altitude.

        Example:
            ```python
            import aprofiles as apro
            # read example file
            path = "examples/data/E-PROFILE/L2_0-20000-001492_A20210909.nc"
            reader = apro.reader.ReadProfiles(path)
            profiles = reader.read()
            # desaturation below 4000m
            profiles.desaturate_below(z=4000., inplace=True)
            profiles.data.attenuated_backscatter_0.desaturated
            True
            ```

            ![Before desaturation](../../assets/images/saturated.png)
            ![After desaturation](../../assets/images/desaturated.png)
        """

        imax = self._get_index_from_altitude_AGL(z)
        unsaturated_data = copy.deepcopy(self.data[var].data)
        for i in range(len(self.data.time.data)):
            unsaturated_data[i, :imax] = abs(unsaturated_data[i, :imax])

        if inplace:
            self.data[var].data = unsaturated_data
            new_profiles_data = self
        else:
            copied_dataset = copy.deepcopy(self)
            copied_dataset.data[var].data = unsaturated_data
            new_profiles_data = copied_dataset

        # add attribute
        new_profiles_data.data[var].attrs["desaturated"] = True
        return new_profiles_data

    def foc(self, method="cloud_base", var="attenuated_backscatter_0", z_snr=2000., min_snr=2., zmin_cloud=200.):
        """
        Calls :meth:`aprofiles.detection.foc.detect_foc()`.
        """
        return apro.detection.foc.detect_foc(self, method, var, z_snr, min_snr, zmin_cloud)
    
    def clouds(self, method: Literal["dec", "vg"]="dec", time_avg=1, zmin=0, thr_noise=5., thr_clouds=4., min_snr=0.0, verbose=False):
        """
        Calls :meth:`aprofiles.detection.clouds.detect_clouds()`.
        """
        return apro.detection.clouds.detect_clouds(self, method, time_avg, zmin, thr_noise, thr_clouds, min_snr, verbose)

    def pbl(self, time_avg=1, zmin=100., zmax=3000., wav_width=200., under_clouds=True, min_snr=2., verbose=False):
        """
        Calls :meth:`aprofiles.detection.pbl.detect_pbl()`.
        """
        return apro.detection.pbl.detect_pbl(self, time_avg, zmin, zmax, wav_width, under_clouds, min_snr, verbose)

    def inversion(self, time_avg=1, zmin=4000., zmax=6000., min_snr=0., under_clouds=False, method="forward", 
        apriori={"lr": 50., "mec": False, "use_cfg": False}, remove_outliers=False, mass_conc=True, mass_conc_method="mortier_2013", verbose=False,
    ):
        """
        Calls :meth:`aprofiles.retrieval.extinction.inversion()` to calculate extinction profiles.
        Calls :meth:`aprofiles.retrieval.mass_conc.mec()` to calculate Mass to Extinction coefficients if `mass_conc` is true (Default).
        """
        apro.retrieval.extinction.inversion(self, time_avg, zmin, zmax, min_snr, under_clouds, method, apriori, remove_outliers, verbose)
        if mass_conc:
            apro.retrieval.mass_conc.concentration_profiles(self, mass_conc_method, apriori)
        return apro

    def plot(
        self, var="attenuated_backscatter_0", datetime=None, zref="agl", zmin=None, zmax=None, vmin=None, vmax=None, log=False,
        show_foc=False, show_pbl=False, show_clouds=False, cmap="coolwarm", show_fig=True, save_fig=None, **kwargs
    ):
        """
        Plotting method.
        Depending on the variable selected, this method will plot an image, a single profile or a time series of the requested variable.
        See also :ref:`Plotting`.

        Args:
            var (str, optional): Variable to be plotted.
            datetime (:class:`numpy.datetime64`, optional): if provided, plot the profile for closest time. If not, plot an image constructed on all profiles.
            zref ({'agl', 'asl'}, optional): Base reference for the altitude axis.
            zmin (float, optional): Minimum altitude AGL (m).
            zmax (float, optional): Maximum altitude AGL (m).
            vmin (float, optional): Minimum value.
            vmax (float, optional): Maximum value.
            log (bool, optional): Use logarithmic scale.
            show_foc (bool, optional): Show fog or condensation retrievals.
            show_pbl (bool, optional): Show PBL height retrievals.
            show_clouds (bool, optional): Show clouds retrievals.
            cmap (str, optional): Matplotlib colormap.
            show_fig (bool, optional): Show Figure.
            save_fig (str, optional): Path of the saved figure.
        """

        # check if var is available
        if var not in list(self.data.data_vars):
            raise ValueError(
                "{} is not a valid variable. \n List of available variables: {}".format(
                    var, list(self.data.data_vars)
                )
            )

        # here, check the dimension. If the variable has only the time dimension, calls timeseries method
        if datetime is None:
            # check dimension of var
            if len(list(self.data[var].dims)) == 2:
                apro.plot.image.plot(self.data, var, zref, zmin, zmax, vmin, vmax, log, show_foc, show_pbl, show_clouds, cmap, show_fig, save_fig)
            else:
                apro.plot.timeseries.plot(self.data, var, show_fig, save_fig, **kwargs)
        else:
            apro.plot.profile.plot(self.data, datetime, var, zref, zmin, zmax, vmin, vmax, log, show_foc, show_pbl, show_clouds, show_fig, save_fig)
    
    def write(self, base_dir=Path('examples', 'data', 'V-Profiles'), verbose=False):
        """
        Calls :meth:`aprofiles.io.write_profiles.write()`.
        """
        apro.io.write_profiles.write(self, base_dir, verbose=verbose)
        


def _main():
    import aprofiles as apro

    path = "examples/data/E-PROFILE/L2_0-20000-001492_A20210909.nc"
    profiles = apro.reader.ReadProfiles(path).read()

    # basic corrections
    profiles.extrapolate_below(z=150., inplace=True)
    profiles.desaturate_below(z=4000., inplace=True)
    
    # detection
    profiles.foc(method="cloud_base", zmin_cloud=200)
    profiles.clouds(zmin=300, thr_noise=5, thr_clouds=4, verbose=True)
    profiles.pbl()
    #profiles.plot(show_foc=True, show_clouds=True, log=True, vmin=1e-2, vmax=1e1)
    
    # inversion
    profiles.inversion(zmin=4000., zmax=6000., method="forward", apriori={"lr":50.},mass_conc=True, verbose=True)
    
    # write results
    profiles.write()

if __name__ == "__main__":
    _main()



# @author Augustin Mortier
# @desc A-Profiles - Standard workflow

import aprofiles as apro
from pathlib import Path
import json
import warnings


def workflow(path, instruments_types, base_dir, CFG, verbose=False):
    apro_reader = apro.reader.ReadProfiles(path)
    profiles = apro_reader.read()

    nans = int(profiles.data.attenuated_backscatter_0.isnull().sum())
    total = profiles.data.attenuated_backscatter_0.size
    if int(profiles.data.attenuated_backscatter_0.isnull().sum())/profiles.data.attenuated_backscatter_0.size>=0.25:
        warnings.warn(f"Error with {path}. attenuated_backscatter_0 variable has {nans}/{total} values as NaNs.")
        return

    # do the rest if instrument_type in the list
    if profiles.data.attrs['instrument_type'] in instruments_types:

        # extrapolation lowest layers
        profiles.extrapolate_below(z=260., inplace=True)
        
        # cloud detection
        if "cloud" in CFG["parameters"][profiles.data.instrument_type]:
            if "method" in CFG["parameters"][profiles.data.instrument_type]["cloud"]:
                    cloud_method = CFG["parameters"][profiles.data.instrument_type]["cloud"]["method"]
            profiles.clouds(cloud_method, time_avg=1, zmin=300., thr_noise=5., thr_clouds=4., verbose=verbose)
        
        # foc and pbl
        profiles.foc(zmin_cloud=300.)        
        profiles.pbl(zmin=300., zmax=3000., under_clouds=False, min_snr=1.5, verbose=verbose)

        # retrievals
        # inversion method selection
        # 1. default: forward, 50sr
        inversion_method = "forward"
        apriori = {
            "lr": 50,
            "mec": False,
            "use_cfg": False
        }
        # 2. if exist, overwrite with CFG["parameters"]
        var_apriori = CFG["parameters"][profiles.data.instrument_type]["inversion"]["apriori"]["var"]
        if profiles.data.instrument_type in CFG["parameters"]:
            if 'cfg' in CFG["parameters"][profiles.data.instrument_type]["inversion"]["apriori"]:
                cfg_path = CFG["parameters"][profiles.data.instrument_type]["inversion"]["apriori"]["cfg"]
                # open config 
                f = open(Path(Path(__file__).parent, '..', '..', cfg_path))
                aer_ifs = json.load(f)
                f.close()
                station_id = f'{profiles._data.attrs["wigos_station_id"]}-{profiles._data.attrs["instrument_id"]}'
                if station_id in aer_ifs:
                    apriori = {
                        var_apriori: aer_ifs[station_id][var_apriori],
                        "use_cfg": True,
                        "cfg": {
                            "data": aer_ifs[station_id]["data"],
                            "use_default": 'False',
                            "attributes": aer_ifs["attributes"],
                            "path": cfg_path
                        } 
                    }
                    if aer_ifs[station_id]["mec"]:
                        apriori["mec"] = aer_ifs[station_id]["mec"]
                else:
                    apriori = {
                        var_apriori: aer_ifs["attributes"]["default"][var_apriori],
                        "use_cfg": True,
                        "cfg": {
                            "use_default": 'True',
                            "attributes": aer_ifs["attributes"],
                            "path": cfg_path
                        }
                    }
                    if aer_ifs["attributes"]["default"]["mec"]:
                        apriori["mec"] = aer_ifs["attributes"]["default"]["mec"]
                    
            if "inversion" in CFG["parameters"][profiles.data.instrument_type]:
                if "method" in CFG["parameters"][profiles.data.instrument_type]["inversion"]:
                    inversion_method = CFG["parameters"][profiles.data.instrument_type]["inversion"]["method"]
        
        profiles.inversion(zmin=4000., zmax=6000., remove_outliers=True, method=inversion_method, apriori=apriori, verbose=verbose)
        profiles.write(base_dir)



# @author Augustin Mortier
# @desc A-Profiles - Planetary Boundary Layer

import numpy as np
from rich.progress import track

from aprofiles import utils


def detect_pbl(
    profiles,
    time_avg=1,
    zmin=100.0,
    zmax=3000.0,
    wav_width=200.0,
    under_clouds=True,
    min_snr=1.0,
    verbose=False,
):
    """Module for *Planetary Boundary Layer Height* detection.
    Detects Planetary Boundary Layer Height between zmin and zmax by looking at the maximum vertical gradient in each individual profiles.

    Args:
        profiles (aprofiles.profiles.ProfilesData): `ProfilesData` instance.
        time_avg (int, optional): in minutes, the time during which we aggregate the profiles before detecting the PBL.
        zmin (float, optional): maximum altitude AGL, in m, for retrieving the PBL.
        zmin (float, optional): minimum altitude AGL, in m, for retrieving the PBL.
        wav_width (float, optional): Width of the wavelet used in the convolution, in m.
        under_clouds (bool, optional): If True, and if clouds detection have been called before, force the PBL to be found below the first cloud detected in the profile.
        min_snr (float, optional): Minimum SNR at the retrieved PBL height required to return a valid value.
        verbose (bool, optional): verbose mode.

    Returns:
        (aprofiles.profiles.ProfilesData): 
            adds the following (xarray.DataArray) to existing (aprofiles.profiles.ProfilesData):
            
            - `'pbl' (time, altitude)`: mask array corresponding to the pbl height.

    Example:
        ```python
        import aprofiles as apro
        # read example file
        path = "examples/data/L2_0-20000-001492_A20210909.nc"
        reader = apro.reader.ReadProfiles(path)
        profiles = reader.read()
        # pbl detection
        profiles.pbl(zmin=100., zmax=3000.)
        # attenuated backscatter image with pbl up to 6km of altitude
        profiles.plot(show_pbl=True, zmax=6000., vmin=1e-2, vmax=1e1, log=True)
        ```

        ![Planetary Boundary Layer Height detection](../../assets/images/pbl.png)
    """

    from scipy.ndimage.filters import uniform_filter1d

    def _detect_pbl_from_rcs(data, zmin, zmax, wav_width, min_snr):
        # detect pbl from range corrected signal between zmin and zmax using convolution with a wavelet..
        """
        from scipy import signal

        #define wavelet with constant width
        npoints = len(data)
        width = wav_width #in meter
        wav = signal.ricker(npoints, width/profiles._get_resolution('altitude'))

        #simple convolution
        convolution = signal.convolve(data, wav, mode='same')

        #the PBL is the maximum of the convolution
        #sets to nan outside of PBL search range
        convolution[0:profiles._get_index_from_altitude_AGL(zmin):] = np.nan
        convolution[profiles._get_index_from_altitude_AGL(zmax):] = np.nan
        i_pbl = np.nanargmax(abs(convolution))
        """
        #-1. check if any valid data
        if np.isnan(data).all():
            return np.nan

        # 0. rolling average
        avg_data = uniform_filter1d(data, size=10)

        # simple gradient
        gradient = np.gradient(avg_data)

        # the PBL is the maximum of the convolution
        # sets to nan outside of PBL search range
        gradient[0: profiles._get_index_from_altitude_AGL(zmin):] = np.nan
        gradient[profiles._get_index_from_altitude_AGL(zmax):] = np.nan
        if not np.isnan(gradient).all():
            i_pbl = np.nanargmin(gradient)
        else:
            return np.nan

        # calculates SNR
        snr = utils.snr_at_iz(data, i_pbl, step=10)
        if snr > min_snr:
            return profiles.data.altitude.data[i_pbl]
        else:
            return np.nan

    # we work on profiles averaged in time to reduce the noise
    rcs = profiles.time_avg(
        time_avg, var="attenuated_backscatter_0", inplace=False
    ).data.attenuated_backscatter_0

    # if under_clouds, check if clouds_bases is available
    if under_clouds and "clouds" in list(profiles.data.data_vars):
        lowest_clouds = profiles._get_lowest_clouds()
    elif under_clouds and "clouds" not in list(profiles.data.data_vars):
        import warnings

        warnings.warn(
            "under_clouds parameter sets to True (defaults value) when the clouds detection has not been applied to ProfileData object."
        )
        lowest_clouds = [np.nan for _ in np.arange(len(profiles.data.time))]
    else:
        lowest_clouds = [np.nan for _ in np.arange(len(profiles.data.time))]

    pbl = []
    for i in (track(range(len(profiles.data.time.data)), description="pbl   ", disable=not verbose)):
        lowest_cloud_agl = lowest_clouds[i] - profiles.data.station_altitude.data
        _zmax = np.nanmin([zmax, lowest_cloud_agl])
        pbl.append(
            _detect_pbl_from_rcs(
                rcs.data[i, :],
                zmin,
                _zmax,
                wav_width,
                min_snr,
            )
        )

    # creates dataarrays
    profiles.data["pbl"] = ("time", pbl)
    profiles.data["pbl"] = profiles.data.pbl.assign_attrs({
        'long_name': "Planetary Boundary Layer Height, ASL",
        'units': 'm',
        'time_avg': time_avg,
        'zmin': zmin,
        'zmax': zmax
        })
    return profiles


def _main():
    import aprofiles as apro

    path = "examples/data/E-PROFILE/L2_0-20000-001492_A20210909.nc"
    
    profiles = apro.reader.ReadProfiles(path).read()

    # basic corrections
    profiles.extrapolate_below(z=150, inplace=True)
    # detection
    profiles.pbl(
        zmin=300, zmax=3000, wav_width=200, under_clouds=False, min_snr=2, verbose=True
    )
    profiles.plot(show_pbl=True, log=True, vmin=1e-2, vmax=1e1)
    # plot single profile
    datetime = np.datetime64("2021-09-09T21:20:00")
    profiles.plot(datetime=datetime, vmin=-1, vmax=10, zmax=12000, show_pbl=True)


if __name__ == "__main__":
    _main()



# @author Augustin Mortier
# @desc A-Profiles - Extinction to Backscatter class

import random
from datetime import date, datetime, time

import numpy as np
import pandas as pd
import xarray as xr

from aprofiles.profiles import ProfilesData
from aprofiles.rayleigh import RayleighData


class ExtinctionToAttenuatedBackscatter:
    """Class for simulating measurements (attenuated backscatter profiles) for different models.

    Attributes:
        `model` ({'null', 'step'}): atmospheric model to be simulated.
        `wavelength` (float): Wavelength of the Rayleigh profile to be computed, in nm.
        `lidar_ratio` (float): Lidar Ratio, in sr.
        `noise` (float): Noise level. The noise is normalized to the maximum extinction value in the profile.

    Example:
        ```python
        # some imports
        import aprofiles as apro
        # simulate rayleigh profiles with a random noise
        simulator = apro.simulator.ExtinctionToAttenuatedBackscatter(
            model = 'step', wavelength = 1064., lidar_ratio = 50., noise = 0.5
        );
        # calls the to_profiles_data method
        sim_profiles = simulator.to_profiles_data()
        # plot modelled extinction
        sim_profiles.plot('extinction_model')
        ```

        ![Simulation of aerosol extinction using a step model](../../assets/images/simulation_step-model.png)
    """    
    # get the right reading class
    def __init__(self, model, wavelength, lidar_ratio, noise):
        self.model = model
        self.wavelength = wavelength
        self.lidar_ratio = lidar_ratio
        self.noise = noise

        # workflow
        ds = self._model_extinction()
        ds = self._simulate_attenuated_backscatter(ds)
        self.data = ds
    
    def to_profiles_data(self: xr.DataArray):
        """
        Method which returns an instance of the (aprofiles.profiles.ProfilesData) class.

        Returns:
            (aprofiles.profiles.ProfilesData):
        """  
        return ProfilesData(self.data)

    def _model_extinction(self):
        """Calculates the extinction coefficient profiles for a given aerosol model (vertical distribution, lidar ratio) at a given wavelength and with a random noise.

        Returns:
            (xr.Dataset):
        """        
        _time = pd.date_range(
            start=datetime.combine(date.today(), time(0, 0, 0)),
            end=datetime.combine(date.today(), time(23, 59, 59)),
            periods=24 * 60 / 5,
        ).tolist()
        altitude = np.arange(15, 15000, 15)

        # extinction models
        extinction_model = []
        # model clear: twice the molecular
        if self.model == "null":
            extinction_profile = np.asarray([0 for z in altitude])
        if self.model == "step":
            extinction_profile = np.asarray([0.1 if z<3000 else 0 for z in altitude])
        if self.model == "aloft":
            #TODO: gaussian in altitude (3 km)
            pass

        # use profile over the whole day
        for t in _time:
            norm_noise = self.noise*np.nanmax(extinction_profile) / altitude[-1]**2
            noise_profile = [norm_noise * random.uniform(-1,1) * z**2 for z in altitude]
            extinction_model.append(np.asarray(extinction_profile + noise_profile))

        ds = xr.Dataset(
            data_vars=dict(
                extinction_model=(["time", "altitude"], extinction_model),
                station_altitude=(0.0),
                station_latitude=(0.0),
                station_longitude=(0.0),
                l0_wavelength=(self.wavelength)
            ),
            coords=dict(time=_time, altitude=altitude),
            attrs=dict(site_location=f"Simulation - [{self.model}]"),
        )
        ds.extinction_model.attrs = {
            "long_name": f"Extinction Coefficient (model) @ {self.wavelength} nm",
            "units": "km-1",
        }
        ds.l0_wavelength.attrs = {
            "units": "nm"
        }
        return ds
    
    def _simulate_attenuated_backscatter(self, ds: xr.Dataset):
        """Calculates the attenuated backscatter measurements for given extinction profiles
        
        Args:
            ds (xr.Dataset): Dataset which contains extinction profiles (`time`, `altitude`)
        
        Returns:
            (xr.Dataset):
        """
        # frequently used variables
        altitude = ds.altitude.data
        time = ds.time.data
        extinction_model = ds.extinction_model.data * 1e-3 #conversion from km-1 to m-1
        vertical_resolution = min(np.diff(altitude))

        # open molecular profile
        rayleigh = RayleighData(altitude, self.wavelength, T0=298, P0=1013)

        # attenuated_backscatter calculation
        attenuated_backscatter = []
        for i, t in enumerate(time):
            total_backscatter = rayleigh.backscatter + np.divide(extinction_model[i,:], self.lidar_ratio)
            total_extinction = rayleigh.extinction + extinction_model[i]
            optical_depth = np.cumsum(total_extinction * vertical_resolution)
            transmission = np.exp(-optical_depth)
            attenuated_backscatter.append(np.asarray([total_backscatter[j] * (transmission[j]**2) for j, z in enumerate(altitude)]))
        # convert list to np array
        attenuated_backscatter = np.asarray(attenuated_backscatter)

        # add attenuated_backscatter as new dataarray
        ds["attenuated_backscatter_0"] = xr.DataArray(
            data=attenuated_backscatter * 1e6, #conversion from m-1.sr-1 to Mm-1.sr-1
            dims=["time", "altitude"],
            attrs=dict(long_name=f"Attenuated Backscatter @ {self.wavelength}nm", units='Mm-1.sr-1'),
        )
        return ds


def _main():
    simulator = ExtinctionToAttenuatedBackscatter(model="step", wavelength=1064.0, lidar_ratio = 50., noise=0.5)
    sim_profiles = simulator.to_profiles_data()
    sim_profiles.plot('extinction_model')

if __name__ == "__main__":
    _main()



# @author Augustin Mortier
# @desc A-Profiles - Mass Concentration

import json

import aprofiles as apro
import numpy as np
from pathlib import Path


def concentration_profiles(profiles, method, apriori):
    """Calculates Mass concentration profiles for different aerosol types

    Args:
        profiles (aprofiles.profiles.ProfilesData): `ProfilesData` instance.
        method (str): Method for calculating MEC. Must be one of {"mortier_2013", "literature"}.
        apriori (dict): Apriori mec value (m2.g-1).

    Returns:
        (aprofiles.profiles.ProfilesData):
            object with additional (xarray.Dataset):

            - `'mass_concentration:<aer_type>'`

    Example:
        Profiles preparation
        ```python
        import aprofiles as apro
        # read example file
        path = "examples/data/L2_0-20000-001492_A20210909.nc"
        reader = apro.reader.ReadProfiles(path)
        profiles = reader.read()
        # extrapolate lowest layers
        profiles.extrapolate_below(z=150, inplace=True)
        ```

        Forward inversion
        ```python
        # aerosol inversion
        profiles.inversion(
            zmin=4000, zmax=6000, remove_outliers=False, method='forward', 
            method_mass_conc='mortier_2013'
        )
        # plot mass concentration profiles for urban particles
        profiles.plot(var='mass_concentration:urban', zmax=6000, vmin=0, vmax=100)
        ```

        ![Mass concentration profiles in the case of urban particles](../../assets/images/mass_conc-urban.png)
    """

    # read aer_properties.json files
    f = open(Path(Path(__file__).parent,'..','config','aer_properties.json'))
    aer_properties = json.load(f)
    f.close()
    # check if the aer_type exist in the json file
    aer_types = aer_properties.keys()

    # get wavelength
    wavelength = float(profiles.data.l0_wavelength.data)
    if profiles.data.l0_wavelength.units != "nm":
        raise ValueError("wavelength units is not `nm`.")

    for aer_type in aer_types:
        # calculates mec
        mec = apro.mec.MECData(aer_type, wavelength, method)

        # compute mass_concentration profile. Use extinction as base.
        mass_concentration = profiles.data.extinction * 1e-3 #conversion from km-1 to m-1
        # mass_concentration = copy.deepcopy(profiles.data.extinction)
        mass_concentration.data = np.divide(mass_concentration, mec.mec)
        # # conversion from g.m-3 to µg.m-3
        mass_concentration.data = mass_concentration.data * 1e6

        # creates dataset with a dataarray for each aer_type
        profiles.data[f"mass_concentration:{aer_type}"] = (('time', 'altitude'), mass_concentration.data)
        profiles.data[f"mass_concentration:{aer_type}"] = profiles.data[f"mass_concentration:{aer_type}"].assign_attrs({
            'long_name': f"Mass concentration [{aer_type.replace('_', ' ')} particles]",
            'units': 'µg.m-3',
            'mec': mec.mec,
        })
    
    # add ifs mec
    if apriori.get("mec"):
        aer_type = "ifs"
        # compute mass_concentration profile. Use extinction as base.
        mass_concentration = profiles.data.extinction * 1e-3 #conversion from km-1 to m-1
        # mass_concentration = copy.deepcopy(profiles.data.extinction)
        mass_concentration.data = np.divide(mass_concentration, apriori["mec"])
        # # conversion from g.m-3 to µg.m-3
        mass_concentration.data = mass_concentration.data * 1e6

        # creates dataset with a dataarray for each aer_type
        profiles.data[f"mass_concentration:{aer_type}"] = (('time', 'altitude'), mass_concentration.data)
        profiles.data[f"mass_concentration:{aer_type}"] = profiles.data[f"mass_concentration:{aer_type}"].assign_attrs({
            'long_name': f"Mass concentration [{aer_type.replace('_', ' ')}]",
            'units': 'µg.m-3',
            'mec': apriori["mec"],
        })
        
    return profiles


def _main():
    import aprofiles as apro

    path = "examples/data/E-PROFILE/L2_0-20000-001492_A20210909.nc"
    profiles = apro.reader.ReadProfiles(path).read()

    # basic corrections
    profiles.extrapolate_below(z=150.0, inplace=True)
    profiles.inversion(verbose=True, mass_conc_method="mortier_2013")
    profiles.plot("mass_concentration:urban", zmax=6000, vmin=0, vmax=100)


if __name__ == "__main__":
    _main()



# @author Augustin Mortier
# @desc A-Profiles - Time Series plot

import matplotlib.pyplot as plt
import pandas as pd

import seaborn as sns

sns.set_theme()


def plot(da, var="aod", show_fig=True, save_fig=None, **kwargs):
    """
    Plot time series of selected variable from (aprofiles.profiles.ProfilesData): object.

    Args:
        da (xarray.DataArray): DataArray
        var (str, optional): Variable of the DataArray to be plotted.
        show_fig (bool, optional): Show Figure.
        save_fig (str, optional): Path of the saved figure.

    Example:
        ``` python
        import aprofiles as apro
        # read example file
        path = "examples/data/L2_0-20000-001492_A20210909.nc"
        reader = apro.reader.ReadProfiles(path)
        profiles = reader.read()
        # retrieve pbl height
        profiles.pbl(zmin=200, zmax=3000)
        # attenuated backscatter image
        profiles.plot(var="pbl" ymin=0., ymax=3000., min_snr=2.)
        ```

        ![Time series of Planetary Boundary Layer height](../../assets/images/time_series.png)
    """

    def __init__(self):
        pass

    # get kwargs
    ymin = kwargs.get("ymin", None)
    ymax = kwargs.get("ymax", None)

    # time
    time = da.time.data

    fig, axs = plt.subplots(1, 1, figsize=(12, 4))

    # plot time series
    plt.plot(time, da[var].data)

    # limit to altitude range
    if ymin is not None or ymax is not None:
        plt.ylim([ymin, ymax])

    # set title and axis labels
    yyyy = pd.to_datetime(da.time.values[0]).year
    mm = pd.to_datetime(da.time.values[0]).month
    dd = pd.to_datetime(da.time.values[0]).day
    latitude = da.station_latitude.data
    longitude = da.station_longitude.data
    altitude = da.station_altitude.data
    station_id = da.attrs["site_location"]
    # title
    plt.title(
        f"{station_id} ({latitude:.2f};{longitude:.2f};{altitude:.1f}m) - {yyyy}/{mm:02}/{dd:02}",
        weight="bold",
    )
    # labels
    plt.xlabel("Time")

    if "units" in list(da[var].attrs) and da[var].units is not None:
        ylabel = f"{da[var].long_name} ({da[var].units})"
    else:
        ylabel = f"{da[var].long_name}"
    plt.ylabel(ylabel)

    plt.tight_layout()
    if save_fig:
        plt.savefig(save_fig)
    if show_fig:
        plt.show()
